import json
import logging
import time
import uuid
from typing import Text, Dict, Any, Optional, List, Set

from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.database import (
    DomainEntity, DomainAction, DomainSlot,
    DomainIntent, Domain, Template)
from rasax.community.database.admin import ChatToken

logger = logging.getLogger(__name__)


class ProjectsService(object):

    def __init__(self, session: Session):
        self.session = session

    def store_domain(self,
                     domain: Dict[Text, Any],
                     project_id: Text,
                     path: Optional[Text] = None,
                     store_templates: bool = False,
                     username: Text = config.default_username
                     ) -> None:
        """Stores a domain object.

        Args:
            domain: The domain as dictionary.
            project_id: The project_id the domain belongs to.
            path: File path of the domain.
            store_templates: Whether or not to store templates.
            username: Username performing this operation.
        """

        entities = _entities(domain)
        actions = _actions(domain)
        slots = _slots(domain)
        intents = _intents(domain)

        templates = []

        if store_templates:
            templates = _templates(domain, username, project_id)

        store_entities_as_slots = (
            domain.get("config", {}).get("store_entities_as_slots", True))

        # Delete any old domain for this project
        old_domain = self._get_domain(project_id)
        if old_domain:

            # migrate existing templates to new domain if present
            if not store_templates:
                templates = _create_templates_from_domain(old_domain)

            self.session.delete(old_domain)
            self.session.commit()

        domain = Domain(project_id=project_id,
                        store_entities_as_slots=store_entities_as_slots,
                        actions=actions, intents=intents, slots=slots,
                        entities=entities, templates=templates)

        if path:
            domain.path = path

        self.session.add(domain)
        self.session.commit()

    def _get_domain(self, project_id: Text = config.project_name
                    ) -> Optional[Domain]:
        return self.session.query(Domain).filter(
            Domain.project_id == project_id).first()

    def dump_domain_in_local_mode(self, filename: Optional[Text] = None):
        """Dump domain to `filename` in yml format."""

        if config.LOCAL_MODE:
            self._dump_domain(filename)

    def _dump_domain(self, filename: Optional[Text] = None):
        domain = self.get_domain()
        if domain:
            if not filename:
                filename = domain.get("path") or config.default_domain_path
            from rasa.core.domain import Domain
            Domain.from_dict(domain).persist_clean(filename)

    def get_domain(self, project_id: Text = config.project_name) -> Dict[Text, Any]:
        """Returns a stored domain.

        Args:
            project_id: The project id of the domain.

        Returns:
            The domain as dictionary.
        """

        domain = self._get_domain(project_id)

        if domain:
            return domain.as_dict()
        else:
            return {}

    def get_domain_id(self, project_id: Text = config.project_name
                      ) -> Optional[int]:
        """Returns the id of the stored domain.

        Args:
            project_id: The project id of the domain.

        Returns:
            The domain id.
        """

        domain = self._get_domain(project_id)

        if domain:
            return domain.id

        return None

    def get_intents_from_domain(self, project_id: Text) -> Set[Text]:
        """Get all intents from the domain.

        Args:
            project_id: The project id of the domain.

        Returns:
            Set of unique intent names.
        """

        intents = self.session.query(DomainIntent.intent).join(
            DomainIntent.domain).filter(Domain.project_id == project_id).all()
        return set([i for (i,) in intents])

    @staticmethod
    def _expiration_timestamp(lifetime: int = 30) -> float:
        """Generate expiration timestamp `lifetime` days from current time."""

        return time.time() + lifetime * 60 * 60 * 24

    def generate_and_save_token(self, lifetime: int = 30) -> ChatToken:
        """Generate and save chat_token to db with `lifetime` in days."""

        token = uuid.uuid4().hex
        expires = self._expiration_timestamp(lifetime)
        chat_token = ChatToken(token=token, expires=int(expires))
        self.session.add(chat_token)
        self.session.commit()

        return chat_token

    def update_token(self,
                     bot_name: Text,
                     description:
                     Text, lifetime: int = 30):
        """Update chat_token by adding name and description, the expiry date is
        set to 30 days from the current date."""

        expires = self._expiration_timestamp(lifetime)
        token = self.get_token()
        token.bot_name = bot_name
        token.description = description
        token.expires = int(expires)
        self.session.commit()

    def has_token_expired(self, chat_token: Text) -> bool:
        """Return True if `chat_token` has expired, or token is not in db.

        Return False otherwise.
        """

        db_token = self.get_token()
        if db_token.token != chat_token:
            return True

        if int(time.time()) > db_token.expires:
            return True

        return False

    def get_token(self):
        """Get chat_token."""
        return self.session.query(ChatToken).first()


def _entities(domain: Dict[Text, Any]) -> List[DomainEntity]:
    entities = domain.get("entities", [])
    return [DomainEntity(entity=e) for e in entities]


def _actions(domain: Dict[Text, Any]) -> List[DomainAction]:
    actions = domain.get("actions", [])
    actions = [DomainAction(action=a, is_form=False) for a in actions]
    forms = domain.get("forms", [])
    return actions + [DomainAction(action=f, is_form=True) for f in forms]


def _slots(domain: Dict[Text, Any]) -> List[DomainSlot]:
    slots = domain.get("slots", {})
    return [
        DomainSlot(
            slot=s,
            auto_fill=v.get("auto_fill", True),
            initial_value=json.dumps(v["initial_value"]) if v.get("initial_value")
            else None,
            type=v.get("type", "rasa.core.slots.UnfeaturizedSlot"),
            values=json.dumps(v["values"]) if v.get("values") else None,)
        for s, v in slots.items()
    ]


def _intents(domain: Dict[Text, Any]) -> List[DomainIntent]:
    intents_raw = domain.get("intents", [])
    intents = []
    for i in intents_raw:
        if isinstance(i, str):
            name = i
            _config = {}
        else:
            name, _config = next(iter(i.items()))

        intents.append(DomainIntent(intent=name,
                                    use_entities=_config.get("use_entities",
                                                             True)))
    return intents


def _create_templates_from_domain(domain: Domain) -> List[Template]:
    return [
        Template(template=template.template,
                 content=template.content,
                 text=template.text,
                 annotator_id=template.annotator_id,
                 annotated_at=template.annotated_at,
                 project_id=template.project_id
                 ) for template in domain.templates
    ]


def _templates(domain: Dict[Text, Any], username: Text, project_id: Text
               ) -> List[Template]:
    templates = domain.get("templates", {})

    return [
        Template(template=template,
                 content=json.dumps(response),
                 text=response.get("text"),
                 annotator_id=username,
                 annotated_at=time.time(),
                 project_id=project_id)
        for template, responses in templates.items() for response in responses
    ]
